from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import httpx


@dataclass(slots=True)
class RouteContext:
    """Context for route and client calls.

    This bundles transport and logging-related options so that class methods
    can construct a single object and pass it down into route functions and
    the HTTP client layer.
    """

    session: httpx.AsyncClient | None = None
    debug_api: bool = False
    debug_num_stacks_to_drop: int = 1
    parent_class: Optional[str] = None
    log_level: Optional[str] = None



"""AWS stream configuration mappings."""

from dataclasses import dataclass

from ..stream_config import StreamConfig_Mapping, register_mapping

__all__ = [
    "AWSAthenaMapping",
    "AmazonAthenaHighBandwidthMapping",
    "AmazonS3AssumeRoleMapping",
]


@register_mapping("aws-athena")
@dataclass
class AWSAthenaMapping(StreamConfig_Mapping):
    """AWS Athena data provider mapping."""

    data_provider_type: str = "aws-athena"
    sql: str = "query"
    database_name: str = "databaseName"
    table_name: str = "tableName"


@register_mapping("amazon-athena-high-bandwidth")
@dataclass
class AmazonAthenaHighBandwidthMapping(StreamConfig_Mapping):
    """Amazon Athena high bandwidth data provider mapping."""

    data_provider_type: str = "amazon-athena-high-bandwidth"
    sql: str = "enteredCustomQuery"
    database_name: str = "databaseName"


@register_mapping("amazon_s3_assumerole")
@dataclass
class AmazonS3AssumeRoleMapping(StreamConfig_Mapping):
    """Amazon S3 assume role data provider mapping."""

    data_provider_type: str = "amazon_s3_assumerole"
    s3_bucket_category: str = "filesDiscovery"



"""Full authentication classes for Domo (username/password)."""

from dataclasses import dataclass, field
from typing import Optional, Union

import httpx
from dc_logger.decorators import LogDecoratorConfig, log_call

from ..base.exceptions import AuthError
from ..client.response import ResponseGetData
from ..utils.logging import DomoEntityExtractor, DomoEntityResultProcessor
from .base import _DomoAuth_Optional, _DomoAuth_Required


class _DomoFullAuth_Required(_DomoAuth_Required, _DomoAuth_Optional):  # noqa: N801
    """Mixin for required parameters for DomoFullAuth.

    This class provides full authentication functionality using username and password
    credentials to obtain session tokens from Domo's product APIs.

    Attributes:
        domo_username (str): Domo username for authentication
        domo_password (str): Domo password for authentication
    """

    def __init__(
        self,
        domo_username: str,
        domo_password: str,
        domo_instance: str,
        token_name: Optional[str] = None,
        token: Optional[str] = None,
        user_id: Optional[str] = None,
        is_valid_token: bool = False,
    ):
        """Initialize full authentication with username and password.

        Args:
            domo_username (str): Domo username for authentication
            domo_password (str): Domo password for authentication
            domo_instance (str): The Domo instance identifier
            token_name (Optional[str]): Name identifier for the token
            token (Optional[str]): Pre-existing authentication token
            user_id (Optional[str]): The authenticated user's ID
            is_valid_token (bool): Whether the current token is valid

        Raises:
            InvalidCredentialsError: If username or password is empty
            InvalidInstanceError: If domo_instance is empty
        """
        if not domo_username:
            raise AuthError(message="Domo username is required.")
        if not domo_password:
            raise AuthError(message="Domo password is required.")

        if not domo_instance:
            raise AuthError(message="Domo instance is required.")

        self.domo_username = domo_username
        self.domo_password = domo_password

        _DomoAuth_Required.__init__(self, domo_instance)
        _DomoAuth_Optional.__init__(
            self, domo_instance, token_name, token, user_id, is_valid_token
        )

    @property
    def auth_header(self) -> dict:
        """Generate the full authentication header specific to product APIs.

        Returns:
            dict: HTTP headers with 'x-domo-authentication' token, or empty dict if no token
        """
        return {"x-domo-authentication": self.token} if self.token else {}

    @log_call(
        level_name="auth",
        config=LogDecoratorConfig(
            entity_extractor=DomoEntityExtractor(),
            result_processor=DomoEntityResultProcessor(),
        ),
    )
    async def get_auth_token(
        self,
        session: httpx.AsyncClient | None = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        return_raw: bool = False,
        **kwargs,
    ) -> Union[str, ResponseGetData]:
        """Retrieve the authentication token from product APIs using the provided credentials.

        This method authenticates with Domo using username and password to obtain
        a session token that can be used for subsequent API calls.

        Args:
            session (httpx.AsyncClient | None): HTTP client session to use
            debug_api (bool): Whether to enable API debugging
            debug_num_stacks_to_drop (int): Number of stack frames to drop for debugging
            return_raw (bool): Whether to return raw ResponseGetData instead of token string
            **kwargs: Additional keyword arguments

        Returns:
            Union[str, ResponseGetData]: Authentication token string or raw response data

        Raises:
            InvalidCredentialsError: If authentication fails or no token is returned
        """
        from ..routes import auth as auth_routes

        res = await auth_routes.get_full_auth(
            auth=None,
            domo_instance=self.domo_instance,
            domo_username=self.domo_username,
            domo_password=self.domo_password,
            session=session,
            debug_api=debug_api,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        if res.is_success and res.response:
            self.is_valid_token = True

            token = str(res.response.get("sessionToken", ""))
            self.token = token
            self.token_name = self.token_name or "full_auth"

        if not self.token:
            raise AuthError(message="Failed to retrieve authentication token")

        return self.token


@dataclass
class DomoFullAuth(
    _DomoFullAuth_Required,
):
    """Full authentication using Domo username and password.

    This class provides authentication using Domo credentials (username and password)
    to obtain session tokens. It's typically used for direct user authentication
    where username/password login is permitted.

    Attributes:
        domo_instance (str): The Domo instance identifier
        domo_username (str): Domo username for authentication
        domo_password (str): Domo password for authentication (not shown in repr)
        token_name (Optional[str]): Name identifier for the token
        token (Optional[str]): The authentication token (not shown in repr)
        user_id (Optional[str]): The authenticated user's ID
        is_valid_token (bool): Whether the current token is valid

    Example:
        >>> auth = DomoFullAuth(
        ...     domo_instance="mycompany",
        ...     domo_username="user@company.com",
        ...     domo_password="secure_password"
        ... )
        >>> token = await auth.get_auth_token()
    """

    domo_instance: str
    domo_username: str
    domo_password: str = field(repr=False)
    token_name: Optional[str] = None
    token: Optional[str] = field(default=None, repr=False)
    user_id: Optional[str] = None
    is_valid_token: bool = False

    def __post_init__(self):
        """Initialize the authentication mixins after dataclass creation."""
        _DomoFullAuth_Required.__init__(
            self,
            domo_username=self.domo_username,
            domo_password=self.domo_password,
            domo_instance=self.domo_instance,
            token_name=self.token_name,
            token=self.token,
            user_id=self.user_id,
            is_valid_token=self.is_valid_token,
        )



"""Google stream configuration mappings."""

from dataclasses import dataclass

from ..stream_config import StreamConfig_Mapping, register_mapping

__all__ = [
    "GoogleSheetsMapping",
    "GoogleSpreadsheetsMapping",
]


@register_mapping("google-sheets")
@dataclass
class GoogleSheetsMapping(StreamConfig_Mapping):
    """Google Sheets data provider mapping."""

    data_provider_type: str = "google-sheets"
    google_sheets_file_name: str = "spreadsheetIDFileName"


@register_mapping("google-spreadsheets")
@dataclass
class GoogleSpreadsheetsMapping(StreamConfig_Mapping):
    """Google Spreadsheets data provider mapping."""

    data_provider_type: str = "google-spreadsheets"
    google_sheets_file_name: str = "spreadsheetIDFileName"



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

import httpx

from ...auth import DomoAuth
from ...base import (
    DomoEntity,
    DomoManager,
    exceptions as dmde,
)
from ...routes import stream as stream_routes
from ...routes.stream import Stream_CRUD_Error, Stream_GET_Error
from ...utils import chunk_execution as dmce
from ...utils.logging import get_colored_logger
from ..subentity.schedule import DomoSchedule
from .stream_config import StreamConfig

__all__ = [
    "DomoStream",
    "DomoStreams",
    # Stream Route Exceptions
    "Stream_GET_Error",
    "Stream_CRUD_Error",
]

logger = get_colored_logger()


@dataclass(eq=False)
class DomoStream(DomoEntity):
    """A class for interacting with a Domo Stream (dataset connector)"""

    id: str
    parent: Any = field(repr=False)  # DomoDataset

    transport_description: str = None
    transport_version: int = None
    update_method: str = None
    data_provider_name: str = None
    data_provider_key: str = None
    account_id: str = None
    account_display_name: str = None
    account_userid: str = None

    has_mapping: bool = False
    configuration: list[StreamConfig] = field(default_factory=list)
    configuration_tables: list[str] = field(default_factory=list)
    configuration_query: str = None

    Schedule: DomoSchedule = None  # DomoDataset_Schedule
    Account: Any = field(
        default=None, repr=False
    )  # DomoAccount - set via get_account()

    def __post_init__(self):
        """Post-initialization to extract schedule if present"""
        self.extract_schedule_from_raw()

    def extract_schedule_from_raw(self):
        """Extract schedule from stream configuration if available"""

        if self.raw:
            self.Schedule = DomoSchedule.from_parent(parent=self, obj=self.raw)

        return self.Schedule

    @classmethod
    def from_parent(cls, parent, stream_id: str = None):
        return cls(
            parent=parent,
            id=stream_id or parent.raw.get("streamId"),
            raw=parent.raw,
            auth=parent.auth,
        )

    @property
    def dataset_id(self) -> str:
        return self.parent.id

    @property
    def entity_type(self):
        return "STREAM"

    @property
    def display_url(self):
        """Generate URL to view this stream in the Domo UI"""
        return f"https://{self.auth.domo_instance}.domo.com/datasources/{self.dataset_id}/details/data/table"

    @classmethod
    def from_dict(cls, auth, obj, parent: Any | None = None, **kwargs):  # DomoDataset
        data_provider = obj.get("dataProvider", {})
        transport = obj.get("transport", {})
        obj.get("dataSource", {})

        account = obj.get("account", {})

        sd = cls(
            auth=auth,
            parent=parent,  # Will be set by caller if needed
            id=obj.get("id") or kwargs.get("stream_id"),
            transport_description=transport.get("description"),
            transport_version=transport.get("version"),
            update_method=obj.get("updateMethod"),
            data_provider_name=data_provider.get("name"),
            data_provider_key=data_provider.get("key"),
            raw=obj,
            **{k: v for k, v in kwargs.items() if k != "stream_id"},
        )

        if account:
            sd.account_id = account.get("id")
            sd.account_display_name = account.get("displayName")
            sd.account_userid = account.get("userId")

        sd.configuration = [
            StreamConfig.from_json(
                obj=c_obj, data_provider_type=data_provider.get("key"), parent_stream=sd
            )
            for c_obj in obj.get("configuration", [])
        ]

        return sd

    def generate_config_rpt(self):
        res = {}

        for config in self.configuration:
            if config.stream_category != "default" and config.stream_category:
                obj = config.to_dict()
                res.update({obj["field"]: obj["value"]})

        return res

    async def refresh(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        is_get_account: bool = True,
        is_suppress_no_account_config: bool = True,
    ):
        # Only refresh if stream has an ID (some datasets don't have streams)
        if not self.id:
            return self

        if is_get_account:
            await self.get_account(
                force_refresh=True,
                debug_api=debug_api,
                session=session,
                is_suppress_no_account_config=is_suppress_no_account_config,
            )

        await super().refresh(
            debug_api=debug_api,
            session=session,
            is_suppress_no_account_config=is_suppress_no_account_config,
        )

        return self

    @classmethod
    async def get_by_id(
        cls,
        auth: DomoAuth,
        stream_id: str,
        return_raw: bool = False,
        debug_num_stacks_to_drop=2,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        is_get_account: bool = True,
        is_suppress_no_account_config: bool = True,
    ):
        """Get a stream by its ID.

        Args:
            auth: Authentication object
            stream_id: Unique stream identifier
            return_raw: Return raw response without processing
            debug_num_stacks_to_drop: Stack frames to drop for debugging
            debug_api: Enable API debugging
            session: HTTP client session
            is_get_account: If True and account_id is present, retrieve full Account object
            is_suppress_no_account_config: If True, suppress errors when account config is not found

        Returns:
            DomoStream instance or ResponseGetData if return_raw=True

        Raises:
            Stream_GET_Error: If stream retrieval fails
        """
        res = await stream_routes.get_stream_by_id(
            auth=auth,
            stream_id=stream_id,
            session=session,
            parent_class=cls.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            debug_api=debug_api,
        )

        if return_raw:
            return res

        stream = cls.from_dict(auth=auth, obj=res.response)

        # Retrieve Account if account_id is present
        if is_get_account and stream.account_id:
            await stream.get_account(
                session=session,
                debug_api=debug_api,
                force_refresh=True,
                is_suppress_no_account_config=is_suppress_no_account_config,
            )
        return stream

    @classmethod
    async def get_entity_by_id(cls, entity_id: str, auth: DomoAuth, **kwargs):
        return await cls.get_by_id(stream_id=entity_id, auth=auth, **kwargs)

    @classmethod
    async def create(
        cls,
        cnfg_body,
        auth: DomoAuth = None,
        session: httpx.AsyncClient | None = None,
        debug_api: bool = False,
    ):
        return await stream_routes.create_stream(
            auth=auth, body=cnfg_body, session=session, debug_api=debug_api
        )

    async def update(
        self,
        cnfg_body,
        session: httpx.AsyncClient | None = None,
        debug_api: bool = False,
    ):
        res = await stream_routes.update_stream(
            auth=self.auth,
            stream_id=self.id,
            body=cnfg_body,
            session=session,
            debug_api=debug_api,
        )
        return res

    async def get_account(
        self,
        session: httpx.AsyncClient | None = None,
        debug_api: bool = False,
        force_refresh: bool = False,
        is_suppress_no_account_config: bool = True,
    ) -> Any | None:  # DomoAccount
        """Retrieve the Account associated with this stream.

        Args:
            session: HTTP client session
            debug_api: Enable API debugging
            force_refresh: If True, refresh even if Account is already set
            is_suppress_no_account_config: If True, suppress errors when account config is not found

        Returns:
            DomoAccount instance or None if no account_id

        Example:
            >>> stream = await DomoStream.get_by_id(auth=auth, stream_id="123")
            >>> account = await stream.get_account()
            >>> print(f"Account: {account.name}")
        """

        if not self.account_id:
            return None

        if self.Account is not None and not force_refresh:
            return self.Account

        from ..DomoAccount import DomoAccount

        try:
            self.Account = await DomoAccount.get_by_id(
                auth=self.auth,
                account_id=self.account_id,
                session=session,
                debug_api=debug_api,
                is_use_default_account_class=False,
                is_suppress_no_config=is_suppress_no_account_config,
            )
        except dmde.DomoError as e:
            if is_suppress_no_account_config:
                await logger.warning(
                    f"Warning: Could not retrieve account {self.account_id}: {e}"
                )
                self.Account = None
            else:
                raise e from e

        return self.Account


@dataclass
class DomoStreams(DomoManager):
    streams: list[DomoStream] = field(default=None)

    async def get(
        self,
        search_dataset_name: str = None,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
    ):
        from ...routes import datacenter as datacenter_routes

        res = await datacenter_routes.search_datasets(
            auth=self.auth,
            entity_type=datacenter_routes.Datacenter_Enum.DATASET.value,
            session=session,
            search_text=search_dataset_name,
            debug_api=debug_api,
            parent_class=self.__class__.__name__,
        )

        self.streams = await dmce.gather_with_concurrency(
            *[
                DomoStream.get_by_id(self.auth, stream_id=obj["streamId"])
                for obj in res.response
            ],
            n=10,
        )

        return self.streams

    async def upsert(
        self,
        cnfg_body,
        match_name=None,
        auth: DomoAuth = None,
        session: httpx.AsyncClient | None = None,
        debug_api: bool = False,
    ):
        from ...routes import datacenter as datacenter_routes

        res = await datacenter_routes.search_datasets(
            auth=auth or self.auth,
            entity_type=datacenter_routes.Datacenter_Enum.DATASET.value,
            session=session,
            search_text=match_name,
            debug_api=debug_api,
            parent_class=self.__class__.__name__,
        )
        datasets = res.response

        existing_ds = next((ds for ds in datasets if ds.name == match_name), None)

        if existing_ds:
            domo_stream = await DomoStream.get_by_id(
                auth=auth, stream_id=existing_ds.stream_id
            )

            return await domo_stream.update(
                cnfg_body,
                session=session,
                debug_api=False,
            )

        else:
            return await DomoStream.create(
                cnfg_body, auth=auth, session=session, debug_api=debug_api
            )



"""Base entity classes for Domo objects with inheritance hierarchy and lineage support.

This module provides the foundational classes for all Domo entities including datasets,
cards, pages, users, and groups. It implements a hierarchical structure with support
for entity relationships, lineage tracking, and federation.

Classes:
    DomoEntity: Base entity with core functionality
    DomoEntity_w_Lineage: Entity with lineage tracking capabilities
    DomoManager: Base class for entity managers
    DomoSubEntity: Entity that belongs to a parent entity
"""

from ..client.auth import DomoAuth
from .base import DomoBase
import abc
from dataclasses import dataclass, field, fields

from typing import Any, Callable, Optional

from ..utils.convert import convert_snake_to_pascal
from ..client import auth as dmda


@dataclass
class DomoEntity(DomoBase):
    """Base class for all Domo entities (datasets, cards, pages, users, etc.).

    Provides core functionality including authentication, unique identification,
    data conversion utilities, and relationship management. All concrete entity
    types should inherit from this class or one of its subclasses.

    Attributes:
        auth: Authentication object for API requests (hidden in repr)
        id: Unique identifier for the entity
        raw: Raw API response data for the entity (hidden in repr)
        Relations: Relationship controller for managing entity relationships

    Example:
        >>> entity = SomeDomoEntity(auth=auth, id="123", raw={})
        >>> entity.display_url()  # Implemented by subclass
        'https://mycompany.domo.com/...'
    """

    auth: dmda.DomoAuth = field(repr=False)
    id: str
    raw: dict = field(repr=False)  # api representation of the class

    Relations: Any = field(repr=False, default=None)

    # logger: Logger = field(repr=False) ## pass global logger

    def __post_init__(self):
        """Initialize entity with relationship controller."""
        from .relationships import DomoRelationshipController

        self.Relations = DomoRelationshipController(auth=self.auth, parent_entity=self)

    @property
    def _name(self) -> str:
        name = getattr(self, "name", None)

        if not name:
            raise NotImplementedError(
                "This property should be implemented by subclasses."
            )
        return name

    def __eq__(self, other) -> bool:
        """Check equality based on entity ID and class type.

        Args:
            other: Object to compare with

        Returns:
            bool: True if both are the same class type with the same ID
        """
        if self.__class__.__name__ != other.__class__.__name__:
            return False

        return self.id == other.id

    def to_dict(self, override_fn: Optional[Callable] = None) -> dict:
        """Convert all dataclass attributes to a dictionary in pascalCase.

        This method is useful for serializing entity data for API requests
        or data export operations.

        Args:
            override_fn (Optional[Callable]): Custom conversion function to override default behavior

        Returns:
            dict: Dictionary with pascalCase keys and corresponding attribute values

        Example:
            >>> entity.to_dict()
            {'entityId': '123', 'displayName': 'My Entity', ...}
        """

        if override_fn:
            return override_fn(self)

        return {
            convert_snake_to_pascal(field.name): getattr(self, field.name)
            for field in fields(self)
        }

    @property
    @abc.abstractmethod
    def entity_type(self) -> str:
        """Get the EntityType for this entity based on its class name."""
        return self.__class__.__name__

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, auth: DomoAuth, obj: dict):
        """Create an entity instance from a dictionary representation.

        This method should be implemented by subclasses to handle the conversion
        from API response dictionaries to entity objects.

        Raises:
            NotImplementedError: Must be implemented by subclasses
        """
        raise NotImplementedError("This method should be implemented by subclasses.")

    @classmethod
    @abc.abstractmethod
    async def get_by_id(cls, auth: DomoAuth, id: str):
        """Fetch an entity by its unique identifier.

        This method should be implemented by subclasses to handle entity-specific
        retrieval logic from the Domo API.

        Args:
            auth (DomoAuth): Authentication object for API requests
            entity_id (str): Unique identifier of the entity to retrieve

        Raises:
            NotImplementedError: Must be implemented by subclasses
        """
        raise NotImplementedError("This method should be implemented by subclasses.")

    @classmethod
    @abc.abstractmethod
    async def get_entity_by_id(cls, auth: dmda.DomoAuth, entity_id: str):
        """Fetch an entity by its ID

        This method should be implemented by subclasses to fetch the specific
        entity type while ensuring lineage tracking is properly initialized.

        Args:
            auth (DomoAuth): Authentication object for API requests
            entity_id (str): Unique identifier of the entity to retrieve

        Raises:
            NotImplementedError: Must be implemented by subclasses
        """
        raise NotImplementedError("This method should be implemented by subclasses.")

    @abc.abstractmethod
    def display_url(self) -> str:
        """Generate the URL to display this entity in the Domo interface.

        This method should return the direct URL to view the entity in Domo's
        web interface, allowing users to navigate directly to the entity.

        Returns:
            str: Complete URL to view the entity in Domo

        Raises:
            NotImplementedError: Must be implemented by subclasses
        """
        raise NotImplementedError("This method should be implemented by subclasses.")


@dataclass
class DomoEntity_w_Lineage(DomoEntity):
    """Entity with lineage tracking capabilities.

    Extends DomoEntity to include lineage tracking functionality,
    enabling entities to track their relationships and dependencies
    within the Domo ecosystem.

    Attributes:
        Lineage: Lineage tracking object for dependency management (hidden in repr)
    """

    Lineage: Any = field(repr=False)

    def __post_init__(self):
        """Initialize lineage tracking after entity creation."""
        from ..classes.subentity import DomoLineage as dmdl

        # Using protected method until public interface is available
        self.Lineage = dmdl.DomoLineage.from_parent(auth=self.auth, parent=self)

        super().__post_init__()


@dataclass
class DomoManager(DomoBase):
    """Base class for entity managers that handle collections of entities.

    Provides the foundation for manager classes that handle operations
    on collections of entities (e.g., DatasetManager, CardManager).

    Attributes:
        auth: Authentication object for API requests (hidden in repr)
    """

    auth: dmda.DomoAuth = field(repr=False)

    @abc.abstractmethod
    async def get(self, *args, **kwargs):
        """Retrieve entities based on provided criteria.

        Must be implemented by subclasses to handle entity-specific
        retrieval and filtering logic.

        Args:
            *args: Positional arguments for entity retrieval
            **kwargs: Keyword arguments for filtering and options

        Raises:
            NotImplementedError: Must be implemented by subclasses
        """
        raise NotImplementedError("This method should be implemented by subclasses.")


@dataclass
class DomoSubEntity(DomoEntity):
    """Base class for entities that belong to a parent entity.

    Handles entities that are sub-components of other entities,
    such as columns in a dataset or slides in a page. Automatically
    inherits authentication and parent references.

    Attributes:
        parent: Reference to the parent entity
        auth: Authentication object (inherited from parent, hidden in repr)
    """

    parent: DomoEntity
    auth: dmda.DomoAuth = field(repr=False)

    @property
    def parent_id(self):
        return self.parent.id

    @property
    def auth(self):
        return self.parent.auth

    @classmethod
    def from_parent(cls, parent: DomoEntity):
        """Create a sub-entity instance from a parent entity.

        Args:
            parent (DomoEntity): The parent entity to derive from

        Returns:
            DomoSubEntity: New sub-entity instance with inherited properties
        """
        return cls(parent=parent)



# Integrations module - automatically imports all integration modules
# Users can import integrations like: from domolibrary2.integrations import Automation

# Import all integration modules
from . import Automation, RoleHierarchy, shortcut_fn

# Define what gets imported with "from domolibrary2.integrations import *"
__all__ = [
    "Automation",
    "RoleHierarchy",
    "shortcut_fn",
]



"""
CodeEngine Core Route Functions

This module provides core functions for managing Domo CodeEngine packages including
retrieval and testing operations.

Functions:
    get_packages: Retrieve all codeengine packages
    get_codeengine_package_by_id: Retrieve a specific package by ID
    get_package_versions: Retrieve all versions of a package
    get_codeengine_package_by_id_and_version: Retrieve a specific package version
    test_package_is_released: Test if a package version is released
    test_package_is_identical: Test if package code is identical to existing

Enums:
    CodeEngine_Package_Parts: Enum for package parts parameter values
"""

__all__ = [
    "CodeEngine_Package_Parts",
    "get_packages",
    "get_codeengine_package_by_id",
    "get_package_versions",
    "get_codeengine_package_by_id_and_version",
    "test_package_is_released",
    "test_package_is_identical",
    "execute_codeengine_function",
]

from enum import Enum
from typing import Optional

import httpx

from ...auth import DomoAuth
from ...base.base import DomoEnumMixin
from ...client import (
    get_data as gd,
    response as rgd,
)
from .exceptions import (
    CodeEngine_FunctionCallError,
    CodeEngine_GET_Error,
)


class CodeEngine_Package_Parts(DomoEnumMixin, Enum):
    """Enum for package parts parameter values."""

    VERSIONS = "versions"
    FUNCTIONS = "functions"
    CODE = "code"


@gd.route_function
async def get_packages(
    auth: DomoAuth,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: Optional[str] = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """
    Retrieve all codeengine packages.

    Args:
        auth: Authentication object
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing package list

    Raises:
        CodeEngine_GET_Error: If package retrieval fails
    """
    url = f"http://{auth.domo_instance}.domo.com/api/codeengine/v2/packages"

    res = await gd.get_data(
        url=url,
        auth=auth,
        method="get",
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=parent_class,
        is_follow_redirects=True,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise CodeEngine_GET_Error(res=res)

    return res


@gd.route_function
async def get_codeengine_package_by_id(
    auth: DomoAuth,
    package_id: str,
    params: Optional[dict] = None,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: Optional[str] = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """
    Retrieve a specific codeengine package by ID.

    Args:
        auth: Authentication object
        package_id: Package identifier
        params: Query parameters (optional, defaults to {"parts": "versions"})
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing package data

    Raises:
        CodeEngine_FunctionCallError: If package_id is not provided
        CodeEngine_GET_Error: If package retrieval fails
    """
    if not package_id:
        raise CodeEngine_FunctionCallError(
            message="Package ID must be provided.",
            auth=auth,
        )

    url = (
        f"https://{auth.domo_instance}.domo.com/api/codeengine/v2/packages/{package_id}"
    )

    params = params or {"parts": "versions"}

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        params=params,
        session=session,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise CodeEngine_GET_Error(entity_id=package_id, res=res)

    return res


@gd.route_function
async def get_package_versions(
    auth: DomoAuth,
    package_id: str,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: Optional[str] = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """
    Retrieve all versions of a codeengine package.

    Each package can have one or many versions.

    Args:
        auth: Authentication object
        package_id: Package identifier
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing package versions

    Raises:
        CodeEngine_FunctionCallError: If package_id is not provided
        CodeEngine_GET_Error: If version retrieval fails
    """
    if not package_id:
        raise CodeEngine_FunctionCallError(
            message="Package ID must be provided.",
            auth=auth,
        )

    url = f"https://{auth.domo_instance}.domo.com/api/codeengine/v2/packages/{package_id}/versions/"

    params = {"parts": "functions,code"}

    res = await gd.get_data(
        url=url,
        method="get",
        auth=auth,
        params=params,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise CodeEngine_GET_Error(entity_id=package_id, res=res)

    return res


@gd.route_function
async def get_codeengine_package_by_id_and_version(
    auth: DomoAuth,
    package_id: str,
    version: str,
    params: Optional[dict] = None,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: Optional[str] = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """
    Retrieve a specific codeengine package by ID and version.

    Args:
        auth: Authentication object
        package_id: Package identifier
        version: Package version
        params: Query parameters (optional, defaults to {"parts": "functions,code"})
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing package version data

    Raises:
        CodeEngine_FunctionCallError: If package_id or version is not provided
        CodeEngine_GET_Error: If package retrieval fails
    """
    if not package_id or not version:
        raise CodeEngine_FunctionCallError(
            message=f"Package ID {package_id or 'not provided'} and version {version or 'not provided'} must be provided.",
            auth=auth,
        )

    url = f"https://{auth.domo_instance}.domo.com/api/codeengine/v2/packages/{package_id}/versions/{version}"

    params = params or {"parts": "functions,code"}

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        params=params,
        session=session,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise CodeEngine_GET_Error(entity_id=f"{package_id}/v{version}", res=res)

    return res


async def test_package_is_released(
    package_id: str,
    version: str,
    auth: DomoAuth,
    existing_package=None,
    params: Optional[dict] = None,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: Optional[str] = None,
) -> bool:
    """
    Test if a package version is already released.

    Args:
        package_id: Package identifier
        version: Package version
        auth: Authentication object
        existing_package: Pre-fetched package data (optional)
        params: Query parameters (optional)
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging

    Returns:
        True if the package is already released, False otherwise
    """
    existing_package = (
        existing_package
        or (
            await get_codeengine_package_by_id_and_version(
                auth=auth,
                package_id=package_id,
                version=version,
                params=params,
                debug_api=debug_api,
                session=session,
                parent_class=parent_class,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            )
        ).response
    )

    return existing_package.get("released", False)


async def test_package_is_identical(
    package_id: str,
    version: str,
    auth: DomoAuth,
    existing_package=None,
    new_package=None,
    new_code=None,
    params: Optional[dict] = None,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: Optional[str] = None,
) -> bool:
    """
    Test if the code in a new package matches the existing package.

    Args:
        package_id: Package identifier
        version: Package version
        auth: Authentication object
        existing_package: Pre-fetched existing package data (optional)
        new_package: New package data to compare (optional)
        new_code: New code to compare (optional)
        params: Query parameters (optional)
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging

    Returns:
        True if the package code is identical, False otherwise
    """
    existing_package = (
        existing_package
        or (
            await get_codeengine_package_by_id(
                auth=auth,
                package_id=package_id,
                params=params,
                debug_api=debug_api,
                session=session,
                parent_class=parent_class,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            )
        ).response
    )

    new_code = new_code or (new_package.get("code") if new_package else None)

    return existing_package.get("code") == new_code


@gd.route_function
async def execute_codeengine_function(
    auth: DomoAuth,
    package_id: str,
    version: str,
    function_name: str,
    input_variables: dict,
    is_get_logs: bool = True,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    session: httpx.AsyncClient | None = None,
):
    url = f"https://{auth.domo_instance}.domo.com/api/codeengine/v2/packages/{package_id}/versions/{version}/functions/{function_name}"

    res = await gd.get_data(
        method="POST",
        url=url,
        auth=auth,
        body={"inputVariables": input_variables, "settings": {"getLogs": is_get_logs}},
        session=session,
        parent_class=parent_class,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise CodeEngine_GET_Error(
            entity_id=f"{package_id}/v{version}/function/{function_name}", res=res
        )

    if not res.response["status"] == "SUCCESS":
        raise CodeEngine_FunctionCallError(
            message=f"Function execution failed with status {res.response['status']}",
            auth=auth,
            res=res,
        )

    response = res.response.pop("result")

    metadata = res.response

    res.response = {**response, "_metadata": metadata}

    return res



"""Other/miscellaneous stream configuration mappings."""

from dataclasses import dataclass

from ..stream_config import StreamConfig_Mapping, register_mapping

__all__ = [
    "AdobeAnalyticsV2Mapping",
    "PostgreSQLMapping",
    "QualtricsMapping",
    "SharePointOnlineMapping",
]


@register_mapping("adobe-analytics-v2")
@dataclass
class AdobeAnalyticsV2Mapping(StreamConfig_Mapping):
    """Adobe Analytics v2 data provider mapping."""

    data_provider_type: str = "adobe-analytics-v2"
    sql: str = "query"
    adobe_report_suite_id: str = "report_suite_id"


@register_mapping("postgresql")
@dataclass
class PostgreSQLMapping(StreamConfig_Mapping):
    """PostgreSQL data provider mapping."""

    data_provider_type: str = "postgresql"
    sql: str = "query"


@register_mapping("qualtrics")
@dataclass
class QualtricsMapping(StreamConfig_Mapping):
    """Qualtrics data provider mapping."""

    data_provider_type: str = "qualtrics"
    qualtrics_survey_id: str = "survey_id"


@register_mapping("sharepointonline")
@dataclass
class SharePointOnlineMapping(StreamConfig_Mapping):
    """SharePoint Online data provider mapping."""

    data_provider_type: str = "sharepointonline"
    src_url: str = "relativeURL"



"""Stream configuration mappings organized by platform.

This package contains stream configuration mapping subclasses organized by
major platform (Snowflake, AWS, Domo, Google, etc.). Each mapping is
automatically registered using the @register_mapping decorator.

To add a new mapping:
1. Choose the appropriate platform file (or create a new one)
2. Define a subclass of StreamConfig_Mapping
3. Decorate it with @register_mapping("provider-name")
4. Import it in this __init__.py file

Example:
    from ..stream_config import StreamConfig_Mapping, register_mapping

    @register_mapping("my-provider")
    @dataclass
    class MyProviderMapping(StreamConfig_Mapping):
        data_provider_type: str = "my-provider"
        sql: str = "query"
"""

# Import all platform-specific mappings to trigger registration
from . import _default, aws, domo, google, other, snowflake

__all__ = [
    "_default",
    "aws",
    "domo",
    "google",
    "snowflake",
    "other",
]



__all__ = [
    "DomoAccount_Config",
    "AccountConfig_UsesOauthError",
    "DomoAccount_NoConfig_OAuthError",
    "AccountConfig_ProviderTypeNotDefinedError",
    "DomoAccount_NoConfig",
    "DomoAccount_Config_AbstractCredential",
    "DomoAccount_Config_DatasetCopy",
    "DomoAccount_Config_DomoAccessToken",
    "DomoAccount_Config_Governance",
    "DomoAccount_Config_AmazonS3",
    "DomoAccount_Config_AmazonS3Advanced",
    "DomoAccount_Config_AwsAthena",
    "DomoAccount_Config_HighBandwidthConnector",
    "DomoAccount_Config_Snowflake",
    "DomoAccount_Config_SnowflakeUnload_V2",
    "DomoAccount_Config_SnowflakeUnloadAdvancedPartition",
    "DomoAccount_Config_SnowflakeWriteback",
    "DomoAccount_Config_SnowflakeUnload",
    "DomoAccount_Config_SnowflakeFederated",
    "DomoAccount_Config_SnowflakeInternalUnload",
    "DomoAccount_Config_SnowflakeKeyPairAuthentication",
    "AccountConfig",
]

from dataclasses import dataclass, field
from enum import Enum
from typing import Any

from ...base.base import DomoEnumMixin
from ...base.entities import DomoBase
from ...base.exceptions import ClassError
from ...utils import (
    DictDot as util_dd,
    convert as dmcv,
)


class AccountConfig_UsesOauthError(ClassError):
    def __init__(self, cls_instance, data_provider_type):
        super().__init__(
            cls_instance=cls_instance,
            message=f"data provider type {data_provider_type} uses OAuth and therefore wouldn't return a Config object",
        )


class AccountConfig_ProviderTypeNotDefinedError(ClassError):
    def __init__(self, cls_instance, data_provider_type):
        super().__init__(
            cls_instance=cls_instance,
            message=f"data provider type {data_provider_type} not defined yet. Extend the AccountConfig class",
        )


@dataclass
class DomoAccount_Config(DomoBase):
    """DomoAccount Config abstract base class"""

    data_provider_type: str
    is_oauth: bool

    allow_external_use: bool = True

    parent: Any = field(repr=False, default=None)  # DomoAccount
    raw: dict = field(repr=False, default=None)  # from api response

    @property
    def allow_external_use_from_raw(self):
        if not self.raw:
            return None

        allow_external_use = self.raw.get("allowExternalUse")

        if isinstance(allow_external_use, str):
            allow_external_use = dmcv.convert_string_to_bool(allow_external_use)

        return allow_external_use

    @classmethod
    def from_parent(cls, parent, **kwargs):
        return cls(parent=parent, **kwargs)

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        """convert accounts API response into a class object"""
        cls(
            parent=parent,
            raw=obj,
            **kwargs,
        )

    def to_dict(
        self,
        obj=None,
        column_filter: list[str] = None,  # enumerate columns to include
        **kwargs,
    ) -> dict:
        s = {"allowExternalUse": self.allow_external_use, **(obj or {}), **kwargs}

        if column_filter:
            return {k: v for k, v in s.items() if k in column_filter}

        return s


@dataclass
class DomoAccount_NoConfig_OAuthError(DomoAccount_Config):
    is_oauth: bool = True

    def __super_init__(self):
        raise AccountConfig_UsesOauthError(
            cls_instance=self, data_provider_type=self.data_provider_type
        )


@dataclass
class DomoAccount_NoConfig(DomoAccount_Config):
    is_oauth: bool = False

    def __super_init__(self):
        raise AccountConfig_ProviderTypeNotDefinedError(
            cls_instance=self, data_provider_type=self.data_provider_type
        )


@dataclass
class DomoAccount_Config_AbstractCredential(DomoAccount_Config):
    credentials: dict = None
    data_provider_type: str = "abstract-credential-store"
    is_oauth: bool = False

    @classmethod
    def from_dict(cls, obj, parent=None, **kwargs):
        return cls.from_parent(
            parent=parent, raw=obj, **{"credentials": obj["credentials"], **kwargs}
        )


@dataclass
class DomoAccount_Config_DatasetCopy(DomoAccount_Config):
    data_provider_type = "dataset-copy"
    is_oauth: bool = False

    domo_instance: str = None
    access_token: str = field(repr=False, default=None)

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        return cls(
            access_token=obj.get("accessToken"),
            domo_instance=obj.get("instance"),
            parent=parent,
            raw=obj,
            data_provider_type="dataset-copy",
        )

    def to_dict(self):
        return super().to_dict(
            {"accessToken": self.access_token, "instance": self.domo_instance}
        )


@dataclass
class DomoAccount_Config_DomoAccessToken(DomoAccount_Config):
    data_provider_type: str = "domo-access-token"
    is_oauth: bool = False

    domo_access_token: str = field(repr=False, default=None)
    username: str = None
    password: str = field(repr=False, default=None)

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        return cls(
            domo_access_token=obj.get("domoAccessToken"),
            username=obj.get("username"),
            password=obj.get("password"),
            parent=parent,
            raw=obj,
            is_oauth=None,
        )

    def to_dict(self):
        return super().to_dict(
            {
                "domoAccessToken": self.domo_access_token or "",
                "username": self.username or "",
                "password": self.password or "",
            }
        )


@dataclass
class DomoAccount_Config_Governance(DomoAccount_Config):
    is_oauth: bool = False
    data_provider_type: str = "domo-governance-d14c2fef-49a8-4898-8ddd-f64998005600"

    domo_instance: str = None
    access_token: str = field(repr=False, default=None)

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        return cls(
            access_token=obj.get("apikey"),
            domo_instance=obj.get("customer"),
            parent=parent,
            raw=obj,
        )

    def to_dict(self):
        return super().to_dict(
            {"apikey": self.access_token, "customer": self.domo_instance}
        )


@dataclass
class DomoAccount_Config_AmazonS3(DomoAccount_Config):
    data_provider_type: str = "amazon-s3"
    is_oauth: bool = False

    access_key: str = None
    secret_key: str = field(repr=False, default=None)
    bucket: str = None
    region: str = "us-west-2"

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        dd = util_dd.DictDot(obj)

        return cls(
            access_key=dd.accessKey,
            secret_key=dd.secretKey,
            bucket=dd.bucket,
            region=dd.region,
            parent=parent,
            raw=obj,
        )

    def to_dict(self):
        bucket = self.bucket

        if bucket and bucket.lower().startswith("s3://"):
            bucket = bucket[5:]
            print(
                f"🤦‍♀️- Domo bucket expects string without s3:// prefix. Trimming to '{bucket}' for the output"
            )
        return super().to_dict(
            {
                "accessKey": self.access_key,
                "secretKey": self.secret_key,
                "bucket": bucket,
                "region": self.region,
            }
        )


@dataclass
class DomoAccount_Config_AmazonS3Advanced(DomoAccount_Config):
    data_provider_type: str = "amazons3-advanced"
    is_oauth: bool = False

    access_key: str = None
    secret_key: str = field(repr=False, default=None)

    bucket: str = None
    region: str = "us-west-2"

    @classmethod
    def from_dict(cls, obj: dict = None, parent: Any = None, **kwargs):
        dd = util_dd.DictDot(obj)

        return cls(
            access_key=dd.accessKey,
            secret_key=dd.secretKey,
            bucket=dd.bucket,
            region=dd.region,
            parent=parent,
        )

    def to_dict(self):
        bucket = self.bucket

        if bucket and bucket.lower().startswith("s3://"):
            bucket = bucket[5:]
            print(
                f"🤦‍♀️- Domo bucket expects string without s3:// prefix. Trimming to '{bucket}' for the output"
            )
        return super().to_dict(
            {
                "accessKey": self.access_key,
                "secretKey": self.secret_key,
                "bucket": bucket,
                "region": self.region,
            }
        )


@dataclass
class DomoAccount_Config_AwsAthena(DomoAccount_Config):
    data_provider_type: str = "aws-athena"
    is_oauth: bool = False

    access_key: str = None
    secret_key: str = field(repr=False, default=None)
    bucket: str = None
    workgroup: str = None

    region: str = "us-west-2"

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        dd = util_dd.DictDot(obj)

        return cls(
            access_key=dd.awsAccessKey,
            secret_key=dd.awsSecretKey,
            bucket=dd.s3StagingDir,
            region=dd.region,
            workgroup=dd.workgroup,
            parent=parent,
            raw=obj,
        )

    def to_dict(self):
        return super().to_dict(
            {
                "awsAccessKey": self.access_key,
                "awsSecretKey": self.secret_key,
                "s3StagingDir": self.bucket,
                "region": self.region,
                "workgroup": self.workgroup,
            }
        )


@dataclass
class DomoAccount_Config_HighBandwidthConnector(DomoAccount_Config):
    """this connector is not enabled by default contact your CSM / AE"""

    data_provider_type: str = "amazon-athena-high-bandwidth"
    is_oauth: bool = False

    access_key: str = None
    secret_key: str = field(repr=False, default=None)
    bucket: str = None

    region: str = "us-west-2"
    workgroup: str = None

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        return cls(
            access_key=obj["awsAccessKey"],
            secret_key=obj["awsSecretKey"],
            bucket=obj["s3StagingDir"],
            region=obj["region"],
            data_provider_type="amazon-athena-high-bandwidth",
            workgroup=obj.get("workgroup"),
            parent=parent,
            raw=obj,
        )

    def to_dict(self):
        return super().to_dict(
            {
                "awsAccessKey": self.access_key,
                "awsSecretKey": self.secret_key,
                "s3StagingDir": self.bucket,
                "region": self.region,
                "workgroup": self.workgroup,
            }
        )


@dataclass
class DomoAccount_Config_Snowflake(DomoAccount_Config):
    """this connector is not enabled by default contact your CSM / AE"""

    data_provider_type: str = "snowflake"
    is_oauth: bool = False

    account: str = None
    username: str = None
    password: str = field(repr=False, default=None)
    role: str = None

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        dd = util_dd.DictDot(obj)

        return cls(
            account=dd.account,
            username=dd.username,
            password=dd.password,
            role=dd.role,
            raw=obj,
            parent=parent,
        )

    def to_dict(self):
        return super().to_dict(
            {
                "account": self.account,
                "username": self.username,
                "password": self.password,
                "role": self.role,
            }
        )


@dataclass
class DomoAccount_Config_SnowflakeUnload_V2(DomoAccount_Config):
    """this connector is not enabled by default contact your CSM / AE"""

    data_provider_type: str = "snowflake-unload-v2"
    is_oauth: bool = False

    account: str = None
    username: str = None
    password: str = field(repr=False, default=None)

    access_key: str = None
    secret_key: str = field(repr=False, default=None)
    region: str = None
    bucket: str = None

    role: str = None

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        dd = util_dd.DictDot(obj)

        return cls(
            account=dd.account,
            username=dd.username,
            password=dd.password,
            access_key=dd.accessKey,
            secret_key=dd.secretKey,
            bucket=dd.bucket,
            region=dd.region,
            role=dd.role,
            raw=obj,
            parent=parent,
        )

    def to_dict(self):
        return super().to_dict(
            {
                "account": self.account,
                "username": self.username,
                "password": self.password,
                "role": self.role,
                "accessKey": self.access_key,
                "secretKey": self.secret_key,
                "bucket": self.bucket,
                "region": self.region,
            }
        )


@dataclass
class DomoAccount_Config_SnowflakeUnloadAdvancedPartition(DomoAccount_Config):
    data_provider_type: str = "snowflake-internal-unload-advanced-partition"
    is_oauth: bool = False

    password: str = field(repr=False, default=None)
    account: str = None
    username: str = None
    role: str = None

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        return cls(
            password=obj["password"],
            role=obj.get("role"),
            account=obj["account"],
            username=obj["username"],
            raw=obj,
            parent=parent,
        )

    def to_dict(self):
        return super().to_dict(
            {
                "password": self.password,
                "role": self.role,
                "account": self.account,
                "username": self.username,
            }
        )


@dataclass
class DomoAccount_Config_SnowflakeWriteback(DomoAccount_Config):
    data_provider_type: str = "snowflake-writeback"
    is_oauth: bool = False

    domo_client_secret: str = field(repr=False, default=None)
    domo_client_id: str = None
    account: str = None
    password: str = field(repr=False, default=None)
    username: str = None

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        return cls(
            domo_client_secret=obj["domoClientSecret"],
            domo_client_id=obj["domoClientId"],
            account=obj["account"],
            username=obj["username"],
            password=obj["password"],
            raw=obj,
            parent=parent,
        )

    def to_dict(self):
        return super().to_dict(
            {
                "domoClientSecret": self.domo_client_secret,
                "password": self.password,
                "domoClientId": self.domo_client_id,
                "account": self.account,
                "username": self.username,
            }
        )


@dataclass
class DomoAccount_Config_SnowflakeUnload(DomoAccount_Config):
    data_provider_type: str = "snowflake-unload"
    is_oauth: bool = False

    secret_key: str = field(repr=False, default=None)
    access_key: str = None
    account: str = None
    password: str = field(repr=False, default=None)
    username: str = None
    bucket: str = None

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        return cls(
            secret_key=obj["secretKey"],
            access_key=obj["accessKey"],
            account=obj["account"],
            username=obj["username"],
            password=obj["password"],
            bucket=obj["bucket"],
            raw=obj,
            parent=parent,
        )

    def to_dict(self):
        return super().to_dict(
            {
                "bucket": self.bucket,
                "password": self.password,
                "secretKey": self.secret_key,
                "accessKey": self.access_key,
                "account": self.account,
                "username": self.username,
            }
        )


@dataclass
class DomoAccount_Config_SnowflakeFederated(DomoAccount_Config):
    data_provider_type: str = "snowflake-federated"
    is_oauth: bool = False
    password: str = field(repr=False, default=None)

    host: str = None
    warehouse: str = None
    username: str = None
    port: str = None
    role: str = None

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        return cls(
            password=obj["password"],
            host=obj["host"],
            warehouse=obj["warehouse"],
            username=obj["user"],
            role=obj.get("role"),
            port=obj.get("port"),
            raw=obj,
            parent=parent,
        )

    def to_dict(self):
        return super().to_dict(
            {
                "password": self.password,
                "port": self.port,
                "host": self.host,
                "warehouse": self.warehouse,
                "user": self.username,
                "role": self.role,
            }
        )


@dataclass
class DomoAccount_Config_SnowflakeInternalUnload(DomoAccount_Config):
    is_oauth: bool = False
    data_provider_type: str = "snowflake-internal-unload"

    password: str = field(repr=False, default=None)
    account: str = None
    username: str = None
    role: str = None

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        return cls(
            password=obj["password"],
            role=obj.get("role"),
            account=obj["account"],
            username=obj["username"],
            parent=parent,
            raw=obj,
        )

    def to_dict(self):
        return super().to_dict(
            {
                "password": self.password,
                "role": self.role,
                "account": self.account,
                "username": self.username,
            }
        )


@dataclass
class DomoAccount_Config_SnowflakeKeyPairAuthentication(DomoAccount_Config):
    data_provider_type: str = "snowflakekeypairauthentication"
    is_oauth: bool = False

    private_key: str = field(repr=False, default=None)
    account: str = field(repr=False, default=None)
    passphrase: str = field(repr=False, default=None)
    username: str = None
    role: str = None

    @classmethod
    def from_dict(cls, obj: dict, parent: Any = None, **kwargs):
        return cls(
            private_key=obj["privateKey"],
            role=obj.get("role"),
            account=obj["account"],
            username=obj["username"],
            passphrase=obj["passPhrase"],
            parent=parent,
            raw=obj,
        )

    def to_dict(self):
        return super().to_dict(
            {
                "privateKey": self.private_key,
                "role": self.role,
                "account": self.account,
                "username": self.username,
                "passPhrase": self.passphrase,
            }
        )


class AccountConfig(DomoEnumMixin, Enum):
    """
    Enum provides appropriate spelling for data_provider_type and config object.
    The name of the enum should correspond with the data_provider_type with hyphens replaced with underscores.
    """

    abstract_credential_store = DomoAccount_Config_AbstractCredential
    dataset_copy = DomoAccount_Config_DatasetCopy
    domo_access_token = DomoAccount_Config_DomoAccessToken
    domo_governance_d14c2fef_49a8_4898_8ddd_f64998005600 = DomoAccount_Config_Governance
    aws_athena = DomoAccount_Config_AwsAthena
    amazon_athena_high_bandwidth = DomoAccount_Config_HighBandwidthConnector
    amazon_s3 = DomoAccount_Config_AmazonS3
    amazons3_advanced = DomoAccount_Config_AmazonS3Advanced

    snowflake = DomoAccount_Config_Snowflake

    snowflake_unload = DomoAccount_Config_SnowflakeUnload
    snowflake_unload_v2 = DomoAccount_Config_SnowflakeUnload_V2

    snowflake_internal_unload_advanced_partition = (
        DomoAccount_Config_SnowflakeUnloadAdvancedPartition
    )

    snowflake_internal_unload = DomoAccount_Config_SnowflakeInternalUnload

    snowflakekeypairauthentication = DomoAccount_Config_SnowflakeKeyPairAuthentication

    snowflake_writeback = DomoAccount_Config_SnowflakeWriteback
    snowflake_federated = DomoAccount_Config_SnowflakeFederated

    @staticmethod
    def generate_alt_search_str(raw_value):
        return raw_value.lower().replace("-", "_")

    @classmethod
    def _missing_(cls, value):
        _uses_oauth = ["google_spreadsheets"]

        alt_search_str = cls.generate_alt_search_str(value)

        config_match = next(
            (member for member in cls if member.name in [value, alt_search_str]),
            None,
        )

        # best case scenario alt_search yields a result
        if config_match:
            return config_match

        # second best case, display_type is an oauth and therefore has no matching config
        oauth_match = next(
            (
                oauth_str
                for oauth_str in _uses_oauth
                if oauth_str in [value, alt_search_str]
            ),
            None,
        )
        if oauth_match:
            print(AccountConfig_UsesOauthError(cls, value))
            return None

        # worst case, unencountered display_type
        print(AccountConfig_ProviderTypeNotDefinedError(cls, value))
        return None



"""Domo stream configuration mappings."""

from dataclasses import dataclass

from ..stream_config import StreamConfig_Mapping, register_mapping

__all__ = [
    "DatasetCopyMapping",
    "DomoCSVMapping",
]


@register_mapping("dataset-copy")
@dataclass
class DatasetCopyMapping(StreamConfig_Mapping):
    """Dataset copy data provider mapping."""

    data_provider_type: str = "dataset-copy"
    src_url: str = "datasourceUrl"


@register_mapping("domo-csv")
@dataclass
class DomoCSVMapping(StreamConfig_Mapping):
    """Domo CSV data provider mapping."""

    data_provider_type: str = "domo-csv"
    src_url: str = "datasourceUrl"



from dataclasses import dataclass, field

import httpx

from ...auth import DomoAuth
from ...base import exceptions as dmde
from ...base.entities import DomoEntity
from ...base.exceptions import RouteError
from ...routes import group as group_routes
from ...routes.group import Group_CRUD_Error, GroupType_Enum
from .membership import DomoMembership_Group

__all__ = ["Group_Class_Error", "DomoGroup", "DomoGroups"]


class Group_Class_Error(dmde.ClassError):
    def __init__(
        self, cls_instance, message: str = None, entity_name=None, entity_id: str = None
    ):
        super().__init__(
            cls_instance=cls_instance,
            message=message,
            entity_id=entity_id,
            entity_name=entity_name,
        )


@dataclass(eq=False)
class DomoGroup(DomoEntity):
    auth: DomoAuth = field(repr=False)
    id: str

    name: str = None
    type: str = None

    is_system: bool = None

    description: str = None
    members_id_ls: list[str] = field(repr=False, default_factory=list)
    owner_id_ls: list[str] = field(repr=False, default_factory=list)

    members_ls: list[dict] = field(repr=False, default_factory=list)
    owner_ls: list[dict] = field(repr=False, default_factory=list)

    custom_attributes: dict = field(default_factory=dict)

    Membership: DomoMembership_Group = field(repr=False, default=None)

    @property
    def entity_type(self) -> str:
        return "group"

    def __post_init__(self):
        self.Membership = DomoMembership_Group.from_parent(parent=self)

        self.is_system = True if self.type == "system" else False

    @classmethod
    def from_dict(cls, auth: DomoAuth, obj: dict):
        return cls(
            auth=auth,
            id=obj.get("id") or obj.get("groupId"),
            name=obj.get("name"),
            description=obj.get("description"),
            type=obj.get("type") or obj.get("groupType"),
            members_id_ls=obj.get("userIds"),
            owner_ls=obj.get("owners"),
            raw=obj,
        )

    @classmethod
    def _from_grouplist_json(cls, auth: DomoAuth, obj: dict):
        return cls(
            auth=auth,
            id=obj.get("groupId"),
            name=obj.get("name"),
            description=obj.get("description"),
            type=obj.get("groupType"),
            owner_ls=obj.get("owners"),
            owner_id_ls=[owner.id for owner in obj.get("owners", [])],
            members_ls=obj.get("groupMembers", []),
            members_id_ls=[member.id for member in obj.get("groupMembers", [])],
            raw=obj,
        )

    @staticmethod
    def _groups_to_domo_group(json_list, auth: DomoAuth) -> list[dict]:
        domo_groups = [DomoGroup.from_dict(auth=auth, obj=obj) for obj in json_list]

        return domo_groups

    def display_url(self):
        return f"https://{self.auth.domo_instance}.domo.com/admin/groups/{self.id}"

    @classmethod
    async def get_by_id(
        cls,
        auth: DomoAuth,
        group_id: str,
        return_raw: bool = False,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
    ):
        res = await group_routes.get_group_by_id(
            auth=auth,
            group_id=group_id,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=cls.__name__,
        )
        if return_raw:
            return res

        dg = cls.from_dict(auth=auth, obj=res.response)

        # await dg.Membership.get_owners()
        # await dg.Membership.get_members() # disabled because causes recursion

        return dg

    @classmethod
    async def get_entity_by_id(cls, entity_id, **kwargs):
        """
        Internal method to get an entity by ID.
        """
        return await cls.get_by_id(auth=cls.auth, group_id=entity_id, **kwargs)

    @classmethod
    async def create_from_name(
        cls: "DomoGroup",
        auth: "DomoAuth",
        group_name: str = None,
        group_type: GroupType_Enum | str = "open",  # use GroupType_Enum
        description: str = None,
        is_include_manage_groups_role: bool = True,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
    ):
        res = await group_routes.create_group(
            auth=auth,
            group_name=group_name,
            group_type=group_type,
            description=description,
            debug_api=debug_api,
            session=session,
            parent_class=cls.__name__,
        )

        domo_group = cls.from_dict(auth=auth, obj=res.response)

        if is_include_manage_groups_role:
            await domo_group.Membership.add_owner_manage_all_groups_role(
                debug_api=debug_api, session=session
            )

        return domo_group

    async def update_metadata(
        self: "DomoGroup",
        auth: "DomoAuth" = None,
        group_name: str = None,
        group_type: str = None,  # use GroupType_Enum
        description: str = None,
        additional_params: dict = None,
        debug_api: bool = False,
        return_raw: bool = False,
        session: httpx.AsyncClient = None,
    ):
        auth = auth or self.auth

        res = None
        try:
            res = await group_routes.update_group(
                auth=auth,
                group_id=self.id,
                group_name=group_name,
                group_type=group_type,
                description=description,
                additional_params=additional_params,
                debug_api=debug_api,
                session=session,
            )

            if return_raw:
                return res

            updated_group = await DomoGroup.get_by_id(auth=auth, group_id=self.id)

            self.name = updated_group.name or self.name
            self.description = updated_group.description or self.description
            self.type = updated_group.type or self.type

        except Group_CRUD_Error as e:
            if group_type != self.type:
                raise Group_Class_Error(
                    cls_instance=self,
                    entity_name=self.name,
                    entity_id=self.id,
                    message=f"probably cannot change group_type to '{group_type}' from current type '{self.type}' consider passing `addtional_parameters`",
                ) from e

            else:
                raise e from e

        return self

    async def delete(
        self: "DomoGroup",
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        res = await group_routes.delete_groups(
            auth=self.auth,
            group_ids=[str(self.id)],
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
            session=session,
        )

        res.parent_class = self.__class__.__name__

        return res


async def create_from_name(
    cls: DomoGroup,
    auth: DomoAuth,
    group_name: str = None,
    group_type: GroupType_Enum | str = "open",  # use GroupType_Enum
    description: str = None,
    is_include_manage_groups_role: bool = True,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):
    res = await group_routes.create_group(
        auth=auth,
        group_name=group_name,
        group_type=group_type,
        description=description,
        debug_api=debug_api,
        session=session,
        parent_class=cls.__name__,
    )

    domo_group = cls.from_dict(auth=auth, obj=res.response)

    if is_include_manage_groups_role:
        await domo_group.Membership.add_owner_manage_all_groups_role(
            debug_api=debug_api, session=session
        )

    return domo_group


async def update_metadata(
    self: DomoGroup,
    auth: DomoAuth = None,
    group_name: str = None,
    group_type: str = None,  # use GroupType_Enum
    description: str = None,
    additional_params: dict = None,
    debug_api: bool = False,
    return_raw: bool = False,
    session: httpx.AsyncClient = None,
):
    auth = auth or self.auth

    res = None
    try:
        res = await group_routes.update_group(
            auth=auth,
            group_id=self.id,
            group_name=group_name,
            group_type=group_type,
            description=description,
            additional_params=additional_params,
            debug_api=debug_api,
            session=session,
        )

        if return_raw:
            return res

        updated_group = await DomoGroup.get_by_id(auth=auth, group_id=self.id)

        self.name = updated_group.name or self.name
        self.description = updated_group.description or self.description
        self.type = updated_group.type or self.type

    except Group_CRUD_Error as e:
        if group_type != self.type:
            raise Group_Class_Error(
                cls_instance=self,
                entity_name=self.name,
                entity_id=self.id,
                message=f"probably cannot change group_type to '{group_type}' from current type '{self.type}' consider passing `addtional_parameters`",
            ) from e

        else:
            raise e from e

    return self


async def delete(
    self: DomoGroup,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    session: httpx.AsyncClient = None,
):
    res = await group_routes.delete_groups(
        auth=self.auth,
        group_ids=[str(self.id)],
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=self.__class__.__name__,
        session=session,
    )

    res.parent_class = self.__class__.__name__

    return res


@dataclass
class DomoGroups:
    auth: DomoAuth = field(repr=False)

    is_hide_system_groups: bool = None

    groups: list[DomoGroup] = None

    @staticmethod
    def _groups_to_domo_group(json_list, auth: DomoAuth):
        return [DomoGroup.from_dict(auth=auth, obj=obj) for obj in json_list]

    async def get_is_system_groups_visible(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
        return_raw: bool = False,
    ):
        res = await group_routes.is_system_groups_visible(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        self.is_hide_system_groups = res.response["value"]

        return self.is_hide_system_groups

    async def toggle_show_system_groups(
        self,
        is_hide_system_groups: bool,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
        return_raw: bool = False,
    ):
        if self.is_hide_system_groups == is_hide_system_groups:
            if debug_api:
                print("no change required")
            return self.is_hide_system_groups

        res = await group_routes.toggle_system_group_visibility(
            auth=self.auth,
            is_hide_system_groups=is_hide_system_groups,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        self.is_hide_system_groups = not res.response["value"]

        return self.is_hide_system_groups

    async def get(
        self,
        is_hide_system_groups: bool = True,
        return_raw: bool = False,
        debug_api: bool = False,
        debug_num_stacks_to_drop: bool = False,
        session: httpx.AsyncClient = None,
    ):
        await self.toggle_show_system_groups(
            is_hide_system_groups=is_hide_system_groups,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        res = await group_routes.get_all_groups(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
        )

        if return_raw:
            return res

        if len(res.response):
            self.groups = self._groups_to_domo_group(
                json_list=res.response, auth=self.auth
            )

        else:
            self.groups = []

        return self.groups

    async def search_by_name(
        self,
        group_name: list[str],
        is_hide_system_groups: bool = None,
        only_allow_one: bool = True,
        return_raw: bool = False,
        debug_api: bool = False,
        debug_num_stacks_to_drop: bool = False,
        session: httpx.AsyncClient = None,
    ) -> (
        DomoGroup | list[DomoGroup]
    ):  # by default returns one DomoGroup, but can return a list of DomoGroups
        domo_groups = await self.get(
            is_hide_system_groups=is_hide_system_groups,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
            return_raw=return_raw,
        )

        if return_raw:
            return domo_groups

        filter_groups = None
        if isinstance(group_name, str):
            filter_groups = [
                dg for dg in domo_groups if dg.name.lower() == group_name.lower()
            ]

        if isinstance(group_name, list):
            filter_groups = [
                dg
                for dg in domo_groups
                if dg.name.lower() in [gname.lower() for gname in group_name]
            ]

        if not filter_groups:
            raise Group_Class_Error(
                cls_instance=self,
                entity_id=self.auth.domo_instance,
                message=f"{len(domo_groups)} retrieved.  unable to find a group matching {group_name}",
            )

        if only_allow_one:
            return filter_groups[0]

        return filter_groups

    async def upsert(
        self: "DomoGroups",
        group_name: str,
        group_type: str = None,  # if create_group, use routes.class.GroupType_Enum
        description: str = None,
        debug_api: bool = False,
        additional_params: dict = None,
        session: httpx.AsyncClient = None,
    ):
        domo_group = None
        try:
            domo_group = await self.search_by_name(
                group_name=group_name, only_allow_one=True, session=session
            )

        except Group_Class_Error:
            return await DomoGroup.create_from_name(
                auth=self.auth,
                group_name=group_name,
                group_type=group_type,
                description=description,
                debug_api=debug_api,
                session=session,
            )

        try:
            await domo_group.update_metadata(
                group_type=group_type,
                description=description,
                debug_api=debug_api,
                session=session,
            )

        except RouteError:
            await group_routes.update_group(
                auth=self.auth,
                group_id=domo_group.id,
                group_name=group_name,
                # group_type=group_type,
                description=description,
                additional_params=additional_params,
                debug_api=debug_api,
                session=session,
            )

        return domo_group


async def get(
    self,
    is_hide_system_groups: bool = True,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop: bool = False,
    session: httpx.AsyncClient = None,
):
    await self.toggle_show_system_groups(
        is_hide_system_groups=is_hide_system_groups,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )

    res = await group_routes.get_all_groups(
        auth=self.auth,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=self.__class__.__name__,
    )

    if return_raw:
        return res

    if len(res.response):
        self.groups = self._groups_to_domo_group(json_list=res.response, auth=self.auth)

    else:
        self.groups = []

    return self.groups


async def search_by_name(
    self,
    group_name: list[str],
    is_hide_system_groups: bool = None,
    only_allow_one: bool = True,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop: bool = False,
    session: httpx.AsyncClient = None,
) -> (
    DomoGroup | list[DomoGroup]
):  # by default returns one DomoGroup, but can return a list of DomoGroups
    domo_groups = await self.get(
        is_hide_system_groups=is_hide_system_groups,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        return_raw=return_raw,
    )

    if return_raw:
        return domo_groups

    filter_groups = None
    if isinstance(group_name, str):
        filter_groups = [
            dg for dg in domo_groups if dg.name.lower() == group_name.lower()
        ]

    if isinstance(group_name, list):
        filter_groups = [
            dg
            for dg in domo_groups
            if dg.name.lower() in [gname.lower() for gname in group_name]
        ]

    if not filter_groups:
        raise Group_Class_Error(
            cls_instance=self,
            entity_id=self.auth.domo_instance,
            message=f"{len(domo_groups)} retrieved.  unable to find a group matching {group_name}",
        )

    if only_allow_one:
        return filter_groups[0]

    return filter_groups


async def upsert(
    self: DomoGroups,
    group_name: str,
    group_type: str = None,  # if create_group, use routes.class.GroupType_Enum
    description: str = None,
    debug_api: bool = False,
    additional_params: dict = None,
    session: httpx.AsyncClient = None,
):
    domo_group = None
    try:
        domo_group = await self.search_by_name(
            group_name=group_name, only_allow_one=True, session=session
        )

    except Group_Class_Error:
        return await DomoGroup.create_from_name(
            auth=self.auth,
            group_name=group_name,
            group_type=group_type,
            description=description,
            debug_api=debug_api,
            session=session,
        )

    try:
        await domo_group.update_metadata(
            group_type=group_type,
            description=description,
            debug_api=debug_api,
            session=session,
        )

    except RouteError:
        await group_routes.update_group(
            auth=self.auth,
            group_id=domo_group.id,
            group_name=group_name,
            # group_type=group_type,
            description=description,
            additional_params=additional_params,
            debug_api=debug_api,
            session=session,
        )

    return domo_group



"""Snowflake stream configuration mappings."""

from dataclasses import dataclass

from ..stream_config import StreamConfig_Mapping, register_mapping

__all__ = [
    "SnowflakeMapping",
    "SnowflakeFederatedMapping",
    "SnowflakeInternalUnloadMapping",
    "SnowflakeKeypairAuthMapping",
    "SnowflakeKeypairInternalManagedUnloadMapping",
    "SnowflakeUnloadV2Mapping",
    "SnowflakeWritebackMapping",
]


@register_mapping("snowflake")
@dataclass
class SnowflakeMapping(StreamConfig_Mapping):
    """Snowflake data provider mapping."""

    data_provider_type: str = "snowflake"
    sql: str = "query"
    warehouse: str = "warehouseName"
    database_name: str = "databaseName"


@register_mapping("snowflake_federated")
@dataclass
class SnowflakeFederatedMapping(StreamConfig_Mapping):
    """Snowflake federated data provider mapping."""

    data_provider_type: str = "snowflake_federated"


@register_mapping("snowflake-internal-unload")
@dataclass
class SnowflakeInternalUnloadMapping(StreamConfig_Mapping):
    """Snowflake internal unload data provider mapping."""

    data_provider_type: str = "snowflake-internal-unload"
    sql: str = "customQuery"
    database_name: str = "databaseName"
    warehouse: str = "warehouseName"


@register_mapping("snowflakekeypairauthentication")
@dataclass
class SnowflakeKeypairAuthMapping(StreamConfig_Mapping):
    """Snowflake keypair authentication data provider mapping."""

    data_provider_type: str = "snowflakekeypairauthentication"
    sql: str = "query"
    database_name: str = "databaseName"
    warehouse: str = "warehouseName"


@register_mapping("snowflake-keypair-internal-managed-unload")
@dataclass
class SnowflakeKeypairInternalManagedUnloadMapping(StreamConfig_Mapping):
    """Snowflake keypair internal managed unload data provider mapping."""

    data_provider_type: str = "snowflake-keypair-internal-managed-unload"
    sql: str = "customQuery"
    database_name: str = "databaseName"
    warehouse: str = "warehouseName"


@register_mapping("snowflake_unload_v2")
@dataclass
class SnowflakeUnloadV2Mapping(StreamConfig_Mapping):
    """Snowflake unload v2 data provider mapping."""

    data_provider_type: str = "snowflake_unload_v2"
    sql: str = "query"
    warehouse: str = "warehouseName"
    database_name: str = "databaseName"


@register_mapping("snowflake-writeback")
@dataclass
class SnowflakeWritebackMapping(StreamConfig_Mapping):
    """Snowflake writeback data provider mapping."""

    data_provider_type: str = "snowflake-writeback"
    table_name: str = "enterTableName"
    database_name: str = "databaseName"
    warehouse: str = "warehouseName"



"""Base entity classes for Domo objects with inheritance hierarchy and lineage support.

This module provides the foundational classes for all Domo entities including datasets,
cards, pages, users, and groups. It implements a hierarchical structure with support
for entity relationships, lineage tracking, and federation.

Classes:
    DomoEntity: Base entity with core functionality
    DomoEntity_w_Lineage: Entity with lineage tracking capabilities
    DomoManager: Base class for entity managers
    DomoSubEntity: Entity that belongs to a parent entity
"""

import abc
from dataclasses import dataclass, field
from typing import Any, Callable, Optional

import httpx

from ..auth.base import DomoAuth
from ..client.context import RouteContext
from .base import DomoBase
from .relationships import DomoRelationshipController


@dataclass
class DomoEntity(DomoBase):
    """Base class for all Domo entities (datasets, cards, pages, users, etc.).

    Provides core functionality including authentication, unique identification,
    data conversion utilities, and relationship management. All concrete entity
    types should inherit from this class or one of its subclasses.

    Attributes:
        auth: Authentication object for API requests (hidden in repr)
        id: Unique identifier for the entity
        raw: Raw API response data for the entity (hidden in repr)
        Relations: Relationship controller for managing entity relationships

    Example:
        >>> entity = SomeDomoEntity(auth=auth, id="123", raw={})
        >>> entity.display_url()  # Implemented by subclass
        'https://mycompany.domo.com/...'
    """

    auth: DomoAuth = field(repr=False)
    id: str
    raw: dict = field(repr=False)  # api representation of the class

    Relations: DomoRelationshipController = field(repr=False, init=False, default=None)

    _default_route_context: RouteContext | None = field(
        init=False, default=None, repr=False
    )

    @property
    def _name(self) -> str:
        name = getattr(self, "name", None)

        if not name:
            raise NotImplementedError(
                "This property should be implemented by subclasses."
            )
        return name

    def __eq__(self, other) -> bool:
        """Check equality based on entity ID.

        Args:
            other: Object to compare with

        Returns:
            bool: True if both are DomoEntity instances with the same ID
        """
        if isinstance(other, DomoEntity):
            return self.id == other.id

        return False

    def __hash__(self) -> int:
        """Return hash based on entity ID.

        This allows entities to be used in sets and as dictionary keys.
        Two entities with the same ID will have the same hash.

        Returns:
            int: Hash of the entity ID
        """
        return hash(self.id)

    def _build_route_context(
        self,
        *,
        session: httpx.AsyncClient | None = None,
        debug_api: bool | None = None,
        log_level: str | None = None,
        debug_num_stacks_to_drop: int | None = None,
    ) -> RouteContext:
        """Construct a RouteContext for route calls.

        Entity and manager methods should use this helper to build the context
        they pass into route functions, so that shared defaults and per-entity
        settings are applied consistently.
        """

        base = self._default_route_context or RouteContext()

        return RouteContext(
            session=session or base.session,
            debug_api=debug_api if debug_api is not None else base.debug_api,
            debug_num_stacks_to_drop=(
                debug_num_stacks_to_drop
                if debug_num_stacks_to_drop is not None
                else base.debug_num_stacks_to_drop
            ),
            parent_class=self.__class__.__name__,
            log_level=log_level or base.log_level,
        )

    def to_dict(
        self, override_fn: Optional[Callable] = None, return_snake_case: bool = False
    ) -> dict:
        """Convert all dataclass attributes to a dictionary in camelCase or snake_case.

        This method is useful for serializing entity data for API requests
        or data export operations.

        Only fields with repr=True are included, plus any properties listed in
        __serialize_properties__.

        Args:
            override_fn (Optional[Callable]): Custom conversion function to override default behavior
            return_snake_case (bool): If True, return keys in snake_case. If False (default), return camelCase.

        Returns:
            dict: Dictionary with camelCase (default) or snake_case keys and corresponding attribute values

        Example:
            >>> entity.to_dict()
            {'id': '123', 'displayName': 'My Entity', 'displayUrl': '...', ...}
            >>> entity.to_dict(return_snake_case=True)
            {'id': '123', 'display_name': 'My Entity', 'display_url': '...', ...}
        """

        # Use parent's implementation which handles repr filtering and __serialize_properties__
        return super().to_dict(
            override_fn=override_fn, return_snake_case=return_snake_case
        )

    @property
    @abc.abstractmethod
    def entity_type(self) -> str:
        """Get the EntityType for this entity based on its class name."""
        return self.__class__.__name__

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, auth: DomoAuth, obj: dict):
        """Create an entity instance from a dictionary representation.

        This method should be implemented by subclasses to handle the conversion
        from API response dictionaries to entity objects.

        Raises:
            NotImplementedError: Must be implemented by subclasses
        """
        raise NotImplementedError("This method should be implemented by subclasses.")

    @classmethod
    @abc.abstractmethod
    async def get_by_id(
        cls,
        auth: DomoAuth,
        id: str,
        debug_num_stacks_to_drop=2,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
    ):
        """Fetch an entity by its unique identifier.

        This method should be implemented by subclasses to handle entity-specific
        retrieval logic from the Domo API.

        Args:
            auth (DomoAuth): Authentication object for API requests
            entity_id (str): Unique identifier of the entity to retrieve

        Raises:
            NotImplementedError: Must be implemented by subclasses
        """
        raise NotImplementedError("This method should be implemented by subclasses.")

    async def refresh(
        self,
        debug_num_stacks_to_drop=2,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
        **kwargs,
    ):
        """Refresh this instance from the API using its id and auth."""
        result = await type(self).get_entity_by_id(
            auth=self.auth,
            entity_id=self.id,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            debug_api=debug_api,
            session=session,
            **kwargs,
        )
        # Spread attributes from result to self
        if isinstance(result, type(self)):
            self.__dict__.update(
                {k: v for k, v in result.__dict__.items() if v is not None}
            )
        return self

    @classmethod
    @abc.abstractmethod
    async def get_entity_by_id(cls, auth: DomoAuth, entity_id: str):
        """Fetch an entity by its ID

        This method should be implemented by subclasses to fetch the specific
        entity type while ensuring lineage tracking is properly initialized.

        Args:
            auth (DomoAuth): Authentication object for API requests
            entity_id (str): Unique identifier of the entity to retrieve

        Raises:
            NotImplementedError: Must be implemented by subclasses
        """
        raise NotImplementedError("This method should be implemented by subclasses.")

    @property
    @abc.abstractmethod
    def display_url(self) -> str:
        """Generate the URL to display this entity in the Domo interface.

        This method should return the direct URL to view the entity in Domo's
        web interface, allowing users to navigate directly to the entity.

        Returns:
            str: Complete URL to view the entity in Domo

        Raises:
            NotImplementedError: Must be implemented by subclasses
        """
        raise NotImplementedError("This method should be implemented by subclasses.")


@dataclass
class DomoEntity_w_Lineage(DomoEntity):
    """Entity with lineage tracking capabilities.

    Extends DomoEntity to include lineage tracking functionality,
    enabling entities to track their relationships and dependencies
    within the Domo ecosystem.

    Attributes:
        Lineage: Lineage tracking object for dependency management (hidden in repr)
    """

    Lineage: Any = field(repr=False, init=False, default=None)

    def __post_init__(self):
        """Initialize lineage tracking after entity creation."""
        from ..classes.subentity.lineage import DomoLineage

        # Using protected method until public interface is available
        self.Lineage = DomoLineage.from_parent(auth=self.auth, parent=self)

    def _initialize_schedule_from_raw(self):
        """Initialize Schedule from raw API data if schedule information is present.

        This helper method checks for schedule-related fields in the raw API response
        and creates a DomoSchedule instance if any are found. The DomoSchedule factory
        method automatically determines the appropriate schedule type (Simple, Cron,
        or Advanced).

        This method should be called from subclass __post_init__ after setting self.raw.

        Returns:
            Optional[DomoSchedule]: Schedule instance if schedule data exists, None otherwise
        """
        from ..classes.subentity.schedule import DomoSchedule

        if self.raw and any(
            key in self.raw
            for key in [
                "scheduleExpression",
                "scheduleStartDate",
                "advancedScheduleJson",
            ]
        ):
            return DomoSchedule.from_dict(self.raw)

        return None


@dataclass
class DomoManager(DomoBase):
    """Base class for entity managers that handle collections of entities.

    Provides the foundation for manager classes that handle operations
    on collections of entities (e.g., DatasetManager, CardManager).

    Attributes:
        auth: Authentication object for API requests (hidden in repr)
    """

    auth: DomoAuth = field(repr=False)

    @abc.abstractmethod
    async def get(self, *args, **kwargs):
        """Retrieve entities based on provided criteria.

        Must be implemented by subclasses to handle entity-specific
        retrieval and filtering logic.

        Args:
            *args: Positional arguments for entity retrieval
            **kwargs: Keyword arguments for filtering and options

        Raises:
            NotImplementedError: Must be implemented by subclasses
        """
        raise NotImplementedError("This method should be implemented by subclasses.")


@dataclass
class DomoSubEntity(DomoBase):
    """Base class for entities that belong to a parent entity.

    Handles entities that are sub-components of other entities,
    such as columns in a dataset or slides in a page. Automatically
    inherits authentication and parent references.

    Attributes:
        parent: Reference to the parent entity
        auth: Authentication object (inherited from parent, hidden in repr)
    """

    parent: DomoEntity = field(repr=False)

    @property
    def parent_id(self):
        return self.parent.id

    @property
    def auth(self):
        return self.parent.auth

    @classmethod
    def from_parent(cls, parent: DomoEntity):
        """Create a sub-entity instance from a parent entity.

        Args:
            parent (DomoEntity): The parent entity to derive from

        Returns:
            DomoSubEntity: New sub-entity instance with inherited properties
        """
        return cls(parent=parent)



"""
Unified relationship system for Domo entities.

Provides a comprehensive relationship modeling system that unifies
all types of entity interactions including access control, membership,
sharing, and other relationship types within the Domo ecosystem.

Classes:
    DomoRelationship: Represents a relationship between two entities
    DomoRelationshipController: Manages relationships for entities
"""

from .base import DomoBase, DomoEnum
from .entities import DomoEntity
from abc import abstractmethod
from ..utils import chunk_execution as dmce
from dataclasses import dataclass, field
from typing import Any, Dict, List

RelationshipType = DomoEnum
"""Types of relationships between Domo entities."""

# Access and Permission Relationships
# OWNER = "owner"
# ADMIN = "admin"
# EDITOR = "editor"
# PARTICIPANT = "participant"
# VIEWER = "viewer"

# # Membership Relationships
# MEMBER = "member"
# OWNER = "owner"

# # LIneage
# PARENT = "parent"
# CHILD = "child"


@dataclass
class DomoRelationship(DomoBase):
    """Represents a relationship between two Domo entities.

    This unified relationship model can represent any type of connection
    between Domo entities including access control, membership, sharing,
    subscriptions, and organizational structures.

    Attributes:
    """

    relative_id: str
    relative_class: DomoEntity = field(repr=False)
    relationship_type: RelationshipType

    # Core relationship identifiers
    parent_entity: DomoEntity = field(repr=False, default=None)
    relative_entity: DomoEntity = field(repr=False, default=None)

    def __eq__(self, other):
        return (
            self.parent_id == other.parent_id
            and self.relationship_type == other.relationship_type
            and self.relative_id == other.relative_id
        )

    async def get_relative(self):
        assert hasattr(
            self.relative_class, "get_entity_by_id"
        ), "relative_class must implement get_entity_by_id method"
        return await self.relative_class.get_entity_by_id(
            entity_id=self.relative_id, auth=self.parent_entity.auth
        )

    @property
    def parent_id(self):
        return self.parent_entity.id

    metadata: Dict[str, Any] = field(default_factory=dict)

    @abstractmethod
    def to_dict(self) -> str:
        """Convert relationship to JSON string."""
        raise NotImplementedError("Subclasses must implement to_dict method.")

    @abstractmethod
    def update(self):
        """Update relationship metadata or properties."""
        raise NotImplementedError("Subclasses must implement update method.")


@dataclass
class DomoRelationshipController(DomoBase):
    """Controller for managing Domo entity relationships.

    This class provides high-level operations for creating, managing, and
    querying relationships between Domo entities. It serves as the primary
    interface for relationship operations.

    will be implemented as DomoDataset.Relations
    with methods Relations.get_owners(), get_members(), add_owners(), add_members()
    """

    relationships: List[DomoRelationship] = field(default_factory=list)

    async def get_relative_entities(self):
        """retrieves relative entities for each relationship
        not the same as getting the parent objects' relationships
        """
        await dmce.gather_with_concurrency(
            *[relationship.get_relative() for relationship in self.relationships], n=10
        )

        return self.relationships

    @abstractmethod
    def add_relationship(
        self,
        relative_id,
        relationship_type: RelationshipType,
    ) -> DomoRelationship:
        """Create a new relationship between entities."""

    @abstractmethod
    def get(
        self,
    ) -> List[DomoRelationship]:
        """Find relationships matching the specified criteria."""
        raise NotImplementedError("Subclasses must implement get method.")


__all__ = [
    "RelationshipType",
    "DomoRelationship",
    "DomoRelationshipController",
]



"""
Base classes and enums for the Domo entity system.

This module provides foundational classes and enhanced enums that serve as
the building blocks for all Domo entities and relationships.
"""

import abc
from dataclasses import dataclass
from enum import Enum


class DomoEnumMixin:
    """Enhanced Enum mixin with case-insensitive lookup and default value support.

    This mixin provides case-insensitive string matching and falls back to a default
    value when no match is found. All subclasses should define a 'default' member.

    Example:
        >>> class Status(DomoEnumMixin, Enum):
        ...     ACTIVE = "active"
        ...     INACTIVE = "inactive"
        ...     default = "UNKNOWN"
        >>> Status.get("ACTIVE")  # Case insensitive
        <Status.ACTIVE: 'active'>
        >>> Status.get("invalid")
        <Status.default: 'UNKNOWN'>
    """

    @classmethod
    def get(cls, value):
        """Get enum member by case-insensitive string lookup.

        Args:
            value: String value to look up (case-insensitive)

        Returns:
            Enum member if found, otherwise the default member
        """
        if not isinstance(value, str):
            return getattr(cls, "default", None)

        # cls should be an Enum subclass at runtime
        for member in cls:  # type: ignore
            if member.name.lower() == value.lower():
                return member

        return getattr(cls, "default", None)

    @classmethod
    def _missing_(cls, value):
        """Handle missing enum values with case-insensitive fallback.

        Args:
            value: The value that wasn't found

        Returns:
            Enum member if case-insensitive match found, otherwise default
        """
        if isinstance(value, str):
            value_lower = value.lower()
            # cls should be an Enum subclass at runtime
            for member in cls:  # type: ignore
                if (
                    hasattr(member, "name")
                    and isinstance(member.name, str)
                    and member.name.lower() == value_lower
                ):
                    return member

        return getattr(cls, "default", None)


class DomoEnum(DomoEnumMixin, Enum):
    default = "UNKNOWN"


@dataclass
class DomoBase(abc.ABC):
    """Abstract base class for all Domo objects.

    This class serves as the foundation for all Domo entities and managers,
    providing a common interface and ensuring consistent implementation
    across the inheritance hierarchy.
    """


__all__ = [
    "DomoEnum",
    "DomoEnumMixin",
    "DomoBase",
]



"""Default stream configuration mapping for unknown provider types."""

from dataclasses import dataclass

from ..stream_config import StreamConfig_Mapping, register_mapping

__all__ = [
    "DefaultMapping",
]


@register_mapping("default")
@dataclass
class DefaultMapping(StreamConfig_Mapping):
    """Default data provider mapping for unknown types."""

    data_provider_type: str = "default"
    is_default: bool = True



from __future__ import annotations

from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import Any

from sqlglot import exp, parse_one

from ...base.base import DomoBase, DomoEnumMixin
from ...routes.stream import Stream_CRUD_Error, Stream_GET_Error

__all__ = [
    "StreamConfig_Mapping",
    "StreamConfig_Mappings",
    "StreamConfig",
    "register_mapping",
    # Route exceptions
    "Stream_GET_Error",
    "Stream_CRUD_Error",
]

# Registry to store mapping classes
_MAPPING_REGISTRY: dict[str, type["StreamConfig_Mapping"]] = {}


def register_mapping(data_provider_type: str):
    """Decorator to register a StreamConfig_Mapping subclass.

    Args:
        data_provider_type: The data provider type identifier (e.g., 'snowflake')

    Example:
        @register_mapping('snowflake')
        @dataclass
        class SnowflakeMapping(StreamConfig_Mapping):
            sql: str = "query"
            warehouse: str = "warehouseName"
            database_name: str = "databaseName"
    """

    def decorator(cls: type[StreamConfig_Mapping]) -> type[StreamConfig_Mapping]:
        _MAPPING_REGISTRY[data_provider_type] = cls
        return cls

    return decorator


@dataclass
class StreamConfig_Mapping(DomoBase):
    data_provider_type: str | None = None

    sql: str = None
    warehouse: str = None
    database_name: str = None
    s3_bucket_category: str = None

    is_default: bool = False

    table_name: str = None
    src_url: str = None
    google_sheets_file_name: str = None
    adobe_report_suite_id: str = None
    qualtrics_survey_id: str = None

    def search_keys_by_value(
        self,
        value_to_search: str,
    ) -> StreamConfig_Mapping | None:
        if self.is_default:
            if value_to_search in ["enteredCustomQuery", "query", "customQuery"]:
                return "sql"

        return next(
            (key for key, value in asdict(self).items() if value == value_to_search),
            None,
        )


# ============================================================================
# Import platform-specific mappings to trigger registration
# ============================================================================

# Import all mappings from the stream_configs subfolder
# This triggers the @register_mapping decorators and populates _MAPPING_REGISTRY
import domolibrary2.classes.DomoDataset.stream_configs  # noqa: E402, F401

# ============================================================================
# StreamConfig_Mappings Enum (Auto-generated from registry)
# ============================================================================


class StreamConfig_Mappings(DomoEnumMixin, Enum):
    """Enum of all registered stream config mappings.

    This enum is automatically populated from the registry created by
    @register_mapping decorators. To add a new mapping, simply create a
    new subclass with the @register_mapping decorator above.
    """

    # Explicit default member to prevent AttributeError
    default = None  # Will be set dynamically via _missing_

    @classmethod
    def _missing_(cls, value):
        """Handle missing enum values by searching the registry."""
        alt_search = value.lower().replace("-", "_")

        # Try direct registry lookup
        if value in _MAPPING_REGISTRY:
            mapping_cls = _MAPPING_REGISTRY[value]
            return cls._create_pseudo_member_(value, mapping_cls())

        # Try normalized search
        for key, mapping_cls in _MAPPING_REGISTRY.items():
            if key.lower().replace("-", "_") == alt_search:
                return cls._create_pseudo_member_(key, mapping_cls())

        # Return default
        return cls.default

    @classmethod
    def search(cls, value, debug_api: bool = False) -> StreamConfig_Mapping:
        """Search for a mapping by data provider type.

        Args:
            value: The data provider type to search for
            debug_api: Whether to print debug messages

        Returns:
            StreamConfig_Mapping instance or default mapping
        """
        alt_search = value.lower().replace("-", "_")

        # Direct registry lookup
        if value in _MAPPING_REGISTRY:
            return _MAPPING_REGISTRY[value]()

        # Normalized search
        for key, mapping_cls in _MAPPING_REGISTRY.items():
            if key.lower().replace("-", "_") == alt_search:
                return mapping_cls()

        # Default fallback
        if debug_api:
            print(f"{value} has not been added to registry, using default")

        return _MAPPING_REGISTRY.get("default", StreamConfig_Mapping)()


@dataclass
class StreamConfig:
    stream_category: str
    name: str
    type: str
    value: str
    value_clean: str = None
    parent: Any = field(repr=False, default=None)

    def __post_init__(self):
        # self.value_clean = self.value.replace("\n", " ")
        # sc.value_clean = re.sub(" +", " ", sc.value_clean)

        if self.stream_category == "sql" and self.parent:
            self.process_sql()

    def process_sql(self):
        if not self.parent:
            return None

        self.parent.configuration_query = self.value

        try:
            for table in parse_one(self.value).find_all(exp.Table):
                self.parent.configuration_tables.append(table.name.lower())
                self.parent.configuration_tables = sorted(
                    list(set(self.parent.configuration_tables))
                )
        except Exception:
            return None

        return self.parent.configuration_tables

    @classmethod
    def from_json(cls, obj: dict, data_provider_type: str, parent_stream: Any = None):
        config_name = obj["name"]

        mapping_enum = StreamConfig_Mappings.search(data_provider_type)

        stream_category = "default"
        if mapping_enum:
            stream_category = mapping_enum.value.search_keys_by_value(config_name)

            if parent_stream:
                parent_stream.has_mapping = True

        return cls(
            stream_category=stream_category,
            name=config_name,
            type=obj["type"],
            value=obj["value"],
            parent=parent_stream,
        )

    def to_dict(self):
        return {"field": self.stream_category, "key": self.name, "value": self.value}



"""
CodeEngine CRUD Route Functions

This module provides CRUD functions for managing Domo CodeEngine packages including
creation, deployment, and update operations.

Functions:
    deploy_codeengine_package: Deploy a specific package version
    create_codeengine_package: Create a new codeengine package
    increment_version: Increment package version number
    upsert_codeengine_package_version: Create or update a package version
    upsert_package: Create or update a package

Classes:
    CodeEnginePackageBuilder: Helper class for building packages (placeholder)
"""

__all__ = [
    "CodeEnginePackageBuilder",
    "deploy_codeengine_package",
    "create_codeengine_package",
    "increment_version",
    "upsert_codeengine_package_version",
    "upsert_package",
]

from typing import Optional

import httpx

from ...auth import DomoAuth
from ...client import (
    get_data as gd,
    response as rgd,
)
from . import core as codeengine_routes
from .exceptions import (
    CodeEngine_CRUD_Error,
    CodeEngine_GET_Error,
    CodeEngine_InvalidPackageError,
)


class CodeEnginePackageBuilder:
    """Helper class for building CodeEngine packages."""

    def __init__(self):
        pass


@gd.route_function
async def deploy_codeengine_package(
    auth: DomoAuth,
    package_id: str,
    version: str,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: Optional[str] = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """
    Deploy a specific codeengine package version.

    Args:
        auth: Authentication object
        package_id: Package identifier
        version: Package version to deploy
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing deployment result

    Raises:
        CodeEngine_CRUD_Error: If deployment fails
    """
    url = f"https://{auth.domo_instance}.domo.com/api/codeengine/v2/packages/{package_id}/versions/{version}/release"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise CodeEngine_CRUD_Error(
            operation="deploy", entity_id=f"{package_id}/v{version}", res=res
        )

    return res


@gd.route_function
async def create_codeengine_package(
    auth: DomoAuth,
    payload: dict,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: Optional[str] = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """
    Create a new codeengine package.

    Args:
        auth: Authentication object
        payload: Package data dictionary
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing created package data

    Raises:
        CodeEngine_CRUD_Error: If package creation fails
    """
    url = f"https://{auth.domo_instance}.domo.com/api/codeengine/v2/packages"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        body=payload,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise CodeEngine_CRUD_Error(operation="create", res=res)

    return res


def increment_version(version: str) -> str:
    """
    Increment the version number.

    Increments the last part of a dot-separated version string.

    Args:
        version: Version string (e.g., "1.0.0")

    Returns:
        Incremented version string (e.g., "1.0.1")
    """
    parts = version.split(".")
    # Increment the last part
    parts[-1] = str(int(parts[-1]) + 1)
    return ".".join(parts)


@gd.route_function
async def upsert_codeengine_package_version(
    auth: DomoAuth,
    payload: dict,
    version: Optional[str] = None,
    auto_increment_version: bool = True,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: Optional[str] = None,
    debug_prn: bool = False,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """
    Create or update a codeengine package version.

    If the package version exists and is released, optionally increment the version.
    If the package code is identical to existing, skip the update.

    Args:
        auth: Authentication object
        payload: Package data dictionary
        version: Package version (optional, defaults to version in payload)
        auto_increment_version: Automatically increment version if deployed
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        debug_prn: Enable debug printing
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing package data

    Raises:
        CodeEngine_InvalidPackageError: If package is already deployed and auto_increment is False
        CodeEngine_CRUD_Error: If package creation/update fails
    """
    package_id = payload.get("id")
    version = version or payload.get("version")

    try:
        existing_pkg = await codeengine_routes.get_codeengine_package_by_id_and_version(
            auth=auth,
            package_id=package_id,
            version=version,
            params={"parts": "code"},
            debug_api=debug_api,
            session=session,
            parent_class=parent_class,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )
        if await codeengine_routes.test_package_is_released(
            existing_package=existing_pkg.response,
            package_id=package_id,
            version=version,
            auth=auth,
        ):
            if not auto_increment_version:
                raise CodeEngine_InvalidPackageError(
                    message=f"Package {package_id} v{version} already deployed",
                    auth=auth,
                )

            version = increment_version(version)

        if await codeengine_routes.test_package_is_identical(
            existing_package=existing_pkg.response,
            new_package=payload,
            package_id=package_id,
            version=version,
            auth=auth,
        ):
            if debug_prn:
                print(f"Package {package_id} v{version} is identical; skipping update.")
            return existing_pkg

    except CodeEngine_GET_Error:
        pass  # Not found, continue to create

    return await create_codeengine_package(
        auth=auth,
        payload=payload,
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        return_raw=return_raw,
    )


@gd.route_function
async def upsert_package(
    auth: DomoAuth,
    payload: dict,
    check_different: bool = True,
    create_new_version: bool = False,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: Optional[str] = None,
    debug_prn: bool = False,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """
    Create or update a codeengine package.

    If the package doesn't exist, create it. If it exists, update the version.

    Args:
        auth: Authentication object
        payload: Package data dictionary
        check_different: Check if package is different before updating
        create_new_version: Create a new version instead of updating existing
        session: HTTP client session (optional)
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Stack frames to drop for debugging
        parent_class: Name of calling class for debugging
        debug_prn: Enable debug printing
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing package data

    Raises:
        CodeEngine_CRUD_Error: If package creation/update fails
    """
    package_id = payload.get("id")

    if not package_id:
        if debug_prn:
            print("No Package ID found, creating new package...")

        return await create_codeengine_package(
            auth=auth,
            payload=payload,
            debug_api=debug_api,
            session=session,
            parent_class=parent_class,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            return_raw=return_raw,
        )

    try:
        await codeengine_routes.get_codeengine_package_by_id(
            auth=auth,
            package_id=package_id,
            debug_api=debug_api,
            session=session,
            parent_class=parent_class,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )
    except CodeEngine_GET_Error:
        return await create_codeengine_package(
            auth=auth,
            payload=payload,
            debug_api=debug_api,
            session=session,
            parent_class=parent_class,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            return_raw=return_raw,
        )

    return await upsert_codeengine_package_version(
        payload=payload,
        auth=auth,
        auto_increment_version=True,
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        debug_prn=debug_prn,
        return_raw=return_raw,
    )
